import textwrap

from .generator import Generator


class ArmaGenerator(Generator):
    def generate_bool(self, name, data):
        if name is None:  # We're in an array, we don't have a name. Don't print a newline at the end
            return 'true' if data else 'false'

        return '{}={};\n'.format(name, 'true' if data else 'false')

    def generate_float(self, name, data):
        if name is None:  # We're in an array, we don't have a name. Don't print a newline at the end
            return '{}'.format(data)

        return '{}={};\n'.format(name, data)

    def generate_int(self, name, data):
        if name is None:  # We're in an array, we don't have a name. Don't print a newline at the end
            return '{}'.format(data)

        return '{}={};\n'.format(name, data)

    def generate_class(self, name, data):
        inner = [self.generate_item(key, val) for key, val in data.items()]
        template = textwrap.dedent('''\
            class {name}
            {{
            {contents}}};
            ''')

        retval = template.format(name=name, contents=self._indent(''.join(inner)))
        return retval

    def generate_array(self, name, data):
        inner = [self.generate_item(None, val) for val in data]
        template_regular = textwrap.dedent('''\
            {name}[]=
            {{
            {contents}
            }};
            ''')

        template_inner = '{{{contents}}}'

        template = template_inner if name is None else template_regular
        indent = self._noindent if name is None else self._indent

        retval = template.format(name=name, contents=indent(', '.join(inner)))
        return retval

    def _escape_string(self, data):
        return data.replace('"', '""')

    def generate_string(self, name, data):
        if name is None:  # We're in an array, we don't have a name. Don't print a newline at the end
            return '"{}"'.format(self._escape_string(data))

        return '{}="{}";\n'.format(name, self._escape_string(data))


def generate(data, **kwargs):
    g = ArmaGenerator(**kwargs)
    return g.generate(data)



import textwrap


class Generator:
    # Implement the following methods for your own generator
    def generate_number(self, name, data):
        if data == int(data):
            return self.generate_int(name, int(data))
        return self.generate_float(name, data)

    def generate_bool(self, name, data):
        raise NotImplemented

    def generate_float(self, name, data):
        raise NotImplemented

    def generate_int(self, name, data):
        raise NotImplemented

    def generate_class(self, name, data):
        raise NotImplemented

    def generate_array(self, name, data):
        raise NotImplemented

    def generate_string(self, name, data):
        raise NotImplemented

    # ============================================

    def __init__(self, indent=4, use_tabs=False):
        self.indent_value = indent
        self.indent_character = '\t' if use_tabs else ' '

    def _indent(self, text):
        return textwrap.indent(text, self.indent_character * self.indent_value)

    def _noindent(self, text):
        return text  # Dummy function to simplify code

    # =============================================

    def generate_item(self, name, data):
        item_type = type(data)

        if issubclass(item_type, bool):
            text = self.generate_bool(name, data)
        elif issubclass(item_type, (float, int)):
            text = self.generate_number(name, data)
        elif issubclass(item_type, dict):
            text = self.generate_class(name, data)
        elif issubclass(item_type, (list, tuple)):
            text = self.generate_array(name, data)
        elif issubclass(item_type, str):
            text = self.generate_string(name, data)
        else:
            raise Exception('Can\'t handle item type: {}'.format(item_type))

        return text

    def generate(self, data):
        text_items = []
        for key, val in data.items():
            text_items.append(self.generate_item(key, val))

        return '\n'.join(text_items)



import string
import sys

from collections import OrderedDict

QUOTE = '"'
SEMICOLON = ';'
EQUALS = '='
CURLY_OPEN = '{'
CURLY_CLOSE = '}'
SQUARE_OPEN = '['
SQUARE_CLOSE = ']'
COMMA = ','
PLUS = '+'
MINUS = '-'
SLASH = '/'
DOLLAR = '$'
ASTERISK = '*'

VALID_NAME_CHAR = string.ascii_letters + string.digits + '_.\\'


class ParseError(RuntimeError):
    pass


class Parser:
    def ensure(self, condition, message='Error'):
        if condition:
            return

        raise ParseError('{} at position {}. Before: {}'.format(
            message, self.currentPosition, self.raw[self.currentPosition:self.currentPosition + 50]))

    def detectComment(self):
        try:
            if self.raw[self.currentPosition] == SLASH:
                if self.raw[self.currentPosition + 1] == SLASH:
                    try:
                        indexOfLinefeed = self.raw.index('\n', self.currentPosition)
                        self.currentPosition = indexOfLinefeed
                    except ValueError:
                        self.currentPosition = len(self.raw)
                elif self.raw[self.currentPosition + 1] == ASTERISK:
                    indexCommentEnd = self.raw.find('*/', self.currentPosition)
                    self.currentPosition = len(self.raw) if indexCommentEnd == -1 else indexCommentEnd + len('*/')
        except IndexError:
            pass

    def next(self):
        self.currentPosition += 1
        self.detectComment()
        return self.current()

    def nextWithoutCommentDetection(self):
        self.currentPosition += 1
        return self.current()

    def current(self):
        try:
            return self.raw[self.currentPosition]
        except IndexError:
            return None

    def weHaveADoubleQuote(self):
        return self.raw[self.currentPosition:self.currentPosition + 2] == '""'

    def weHaveAStringLineBreak(self):
        return self.raw[self.currentPosition:self.currentPosition + 6] == '" \\n "'

    def forwardToNextQuote(self):
        try:
            self.currentPosition = self.raw.index(QUOTE, self.currentPosition + 1)
        except ValueError:
            self.currentPosition = len(self.raw)

    def indexOfOrMaxSize(self, haystack, needle, fromPos):
        try:
            return haystack.index(needle, fromPos)
        except ValueError:
            return sys.maxsize

    def parseString(self):
        result = ''
        self.ensure(self.current() == QUOTE)
        self.nextWithoutCommentDetection()
        while True:
            if self.weHaveADoubleQuote():
                result += self.current()
                self.nextWithoutCommentDetection()
            elif self.weHaveAStringLineBreak():
                result += '\n'
                self.next()
                self.forwardToNextQuote()
            elif self.current() == QUOTE:
                break
            else:
                try:
                    result += self.current()
                except TypeError:
                    if self.current() is None:
                        raise ParseError('Got EOF while parsing a string')
                    raise

            self.nextWithoutCommentDetection()

        self.ensure(self.current() == QUOTE)
        self.nextWithoutCommentDetection()
        return result

    def guessExpression(self, s):
        s = s.strip()

        if s[:4].lower() == 'true':
            return True
        elif s[:5].lower() == 'false':
            return False
        elif s.startswith('0x'):
            return int(s, 16)
        elif '.' in s:
            try:
                return float(s)
            except ValueError:
                return s
        else:
            try:
                return int(s)
            except ValueError:
                return s

    def parseUnknownExpression(self):
        posOfExpressionEnd = min(
            self.indexOfOrMaxSize(self.raw, SEMICOLON, self.currentPosition),
            self.indexOfOrMaxSize(self.raw, CURLY_CLOSE, self.currentPosition),
            self.indexOfOrMaxSize(self.raw, COMMA, self.currentPosition)
        )

        expression = self.raw[self.currentPosition:posOfExpressionEnd]
        self.ensure(posOfExpressionEnd != sys.maxsize)
        self.currentPosition = posOfExpressionEnd

        return self.guessExpression(expression)

    def parseNonArrayPropertyValue(self):
        current = self.current()
        if current == CURLY_OPEN:
            return self.parseArray()
        elif current == QUOTE:
            return self.parseString()
        elif current == DOLLAR:
            return self.parseTranslationString()
        else:
            return self.parseUnknownExpression()

    def isValidVarnameChar(self, char):
        return char and char in VALID_NAME_CHAR

    def parsePropertyName(self):
        result = self.current()
        while(self.isValidVarnameChar(self.next())):
            result += self.current()

        return result

    def parseClassValue(self):
        result = self.dict()

        self.ensure(self.current() == CURLY_OPEN)
        self.next()
        self.parseWhitespace()

        while(self.current() != CURLY_CLOSE):
            self.parseProperty(result)
            self.parseWhitespace()

        self.next()

        return result

    def parseArray(self):
        result = []
        self.ensure(self.current() == CURLY_OPEN)
        self.next()
        self.parseWhitespace()

        while self.current() != CURLY_CLOSE:
            result.append(self.parseNonArrayPropertyValue())
            self.parseWhitespace()

            if self.current() == COMMA:
                self.next()
                self.parseWhitespace()
            else:
                break

        self.next()
        return result

    def parseWhitespace(self):
        try:
            while self.isWhitespace():
                self.next()
        except IndexError:
            pass

    def isWhitespace(self):
        return self.raw[self.currentPosition] in ' \t\r\n' or ord(self.raw[self.currentPosition]) < 32

    def parseProperty(self, context):
        value = None
        name = self.parsePropertyName()

        self.parseWhitespace()

        if name == 'class':
            name = self.parsePropertyName()
            self.parseWhitespace()

            if self.current() == ':':
                self.next()
                self.parseWhitespace()
                self.parsePropertyName()
                self.parseWhitespace()

        elif name == 'delete':
            self.parsePropertyName()
            self.parseWhitespace()
            self.ensure(self.current() == SEMICOLON)
            self.next()
            return

        elif name == 'import':
            self.parsePropertyName()
            self.parseWhitespace()
            self.ensure(self.current() == SEMICOLON)
            self.next()
            return

        current = self.current()

        if current == SQUARE_OPEN:
            self.ensure(self.next() == SQUARE_CLOSE)
            self.next()
            self.parseWhitespace()

            self.ensure(self.current() == EQUALS or self.current() == PLUS)
            if self.current() == PLUS:
                self.ensure(self.next() == EQUALS)

            self.next()
            self.parseWhitespace()

            value = self.parseArray()

        elif current == EQUALS:
            self.next()
            self.parseWhitespace()
            value = self.parseNonArrayPropertyValue()

        elif current == CURLY_OPEN:
            value = self.parseClassValue()

        elif current == SLASH:
            if self.next() == SLASH:
                try:
                    self.currentPosition = self.raw.index('\n', self.currentPosition)
                except ValueError:
                    self.currentPosition = len(self.raw)

            else:
                raise ParseError('Unexpected value at pos {}'.format(self.currentPosition))

        else:
            raise ParseError('Unexpected value at pos {}'.format(self.currentPosition))

        context[name] = value

        self.parseWhitespace()
        self.ensure(self.current() == SEMICOLON)
        self.next()

    def translateString(self, txt: str) -> str:
        try:
            return self.translations[txt]
        except KeyError:
            return txt

    def parseTranslationString(self):
        result = ''
        assert self.current() == DOLLAR
        self.next()

        if self.raw[self.currentPosition: self.currentPosition + 3] != 'STR':
            raise ParseError('Invalid translation string beginning')

        while self.current():
            current = self.current()
            if current in (SEMICOLON, COMMA, CURLY_CLOSE):
                break
            else:
                if self.isWhitespace():
                    self.parseWhitespace()
                    break
                else:
                    result += current
            self.nextWithoutCommentDetection()

        if self.current() not in (SEMICOLON, COMMA, CURLY_CLOSE):
            raise ParseError('Syntax error next translation string')

        return self.translateString(result)

    def parse(self, raw, keep_order, translations):
        self.currentPosition = 0
        self.raw = raw
        self.translations = translations or {}
        if keep_order:
            self.dict = OrderedDict
        else:
            self.dict = dict

        result = self.dict()

        self.detectComment()
        self.parseWhitespace()
        while self.current():
            self.parseProperty(result)
            self.parseWhitespace()

        return result


def parse(raw, *, keep_order=False, translations=None):
    p = Parser()
    return p.parse(raw, keep_order, translations)



from setuptools import setup

setup(
    name='armaclass',
    version='0.2.1',
    packages=['armaclass'],
    url='https://github.com/overfl0/Armaclass',
    license='MIT',
    author='Lukasz Taczuk',
    author_email='',
    description='Python parser and generator for Arma class definitions (e.g. sqm files)',
    keywords='arma pbo sqm class parser generator',

    classifiers=[
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',

        'License :: OSI Approved :: MIT License',
    ],
)

# Install in "editable mode" for development:
# pip install -e .
